from __future__ import annotations

import dataclasses
import logging

import pandas as pd
from imblearn.over_sampling import ADASYN, SMOTE, BorderlineSMOTE
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler

from autorad.config import config
from autorad.data.dataset import TrainingData, TrainingInput, TrainingLabels
from autorad.feature_selection.selector import create_feature_selector

log = logging.getLogger(__name__)


def get_not_none_kwargs(**kwargs):
    return {k: v for k, v in kwargs.items() if v is not None}


class Preprocessor:
    def __init__(
        self,
        normalize: bool = True,
        feature_selection_method: str | None = None,
        n_features: int | None = None,
        oversampling_method: str | None = None,
        random_state: int = config.SEED,
    ):
        """Performs preprocessing, including:
        1. normalization
        2. feature selection
        3. oversampling

        Args:
            normalize: whether to normalize to range (0, 1)
            feature_selection_method: algorithm to select key features,
                if None, select all features
            n_features: number of features to select, only applicable to selected
                feature selection methods (see feature_selection.selector)
            oversampling_method: minority class oversampling method,
                if None, no oversampling
            random_state: seed
        """
        self.normalize = normalize
        self.feature_selection_method = feature_selection_method
        self.n_features = n_features
        self.oversampling_method = oversampling_method
        self.random_state = random_state
        self.pipeline = self._build_pipeline()
        self.selected_features = None

    def transform(self, X: pd.DataFrame):
        result_array = self.pipeline.transform(X)
        result_df = pd.DataFrame(result_array, columns=self.selected_features)
        return result_df

    def fit_transform(self, data: TrainingData):
        # copy data
        _data = dataclasses.replace(data)
        X, y = _data.X, _data.y
        result_X = {}
        result_y = {}
        all_features = X.train.columns.tolist()
        X_train_trans, y_train_trans = self.pipeline.fit_transform(
            X.train, y.train
        )
        self.selected_features = self.pipeline["select"].selected_features(
            column_names=all_features
        )
        result_X["train"] = pd.DataFrame(
            X_train_trans, columns=self.selected_features
        )
        result_y["train"] = pd.Series(y_train_trans)
        X_test_trans = self.pipeline.transform(X.test)
        result_X["test"] = pd.DataFrame(
            X_test_trans, columns=self.selected_features
        )
        result_y["test"] = y.test
        if X.val is not None:
            X_val_trans = self.pipeline.transform(X.val)
            result_X["val"] = pd.DataFrame(
                X_val_trans, columns=self.selected_features
            )
            result_y["val"] = y.val
        if X.train_folds is not None and X.val_folds is not None:
            (
                result_X["train_folds"],
                result_y["train_folds"],
                result_X["val_folds"],
                result_y["val_folds"],
            ) = self._fit_transform_cv_folds(_data)
        _data._X_preprocessed = TrainingInput(**result_X)
        _data._y_preprocessed = TrainingLabels(**result_y)
        return _data

    def _fit_transform_cv_folds(
        self, data: TrainingData
    ) -> tuple[
        list[pd.DataFrame],
        list[pd.Series],
        list[pd.DataFrame],
        list[pd.Series],
    ]:
        if (
            data.X.train_folds is None
            or data.y.train_folds is None
            or data.X.val_folds is None
            or data.y.val_folds is None
        ):
            raise AttributeError("Folds are not set")
        (
            result_X_train_folds,
            result_y_train_folds,
            result_X_val_folds,
            result_y_val_folds,
        ) = ([], [], [], [])
        for X_train, y_train, X_val in zip(
            data.X.train_folds,
            data.y.train_folds,
            data.X.val_folds,
        ):
            cv_pipeline = self._build_pipeline()
            all_features = X_train.columns.tolist()
            result_X_train, result_y_train = cv_pipeline.fit_transform(
                X_train, y_train
            )
            selected_features = cv_pipeline["select"].selected_features(
                column_names=all_features
            )
            result_X_val = cv_pipeline.transform(X_val)
            result_df_X_train = pd.DataFrame(
                result_X_train, columns=selected_features
            )
            result_df_X_val = pd.DataFrame(
                result_X_val, columns=selected_features
            )
            result_X_train_folds.append(result_df_X_train)
            result_y_train_folds.append(pd.Series(result_y_train))
            result_X_val_folds.append(result_df_X_val)
        result_y_val_folds = data.y.val_folds
        return (
            result_X_train_folds,
            result_y_train_folds,
            result_X_val_folds,
            result_y_val_folds,
        )

    def _build_pipeline(self):
        steps = []
        if self.normalize:
            steps.append(("normalize", MinMaxWrapper()))
        if self.feature_selection_method is not None:
            steps.append(
                (
                    "select",
                    create_feature_selector(
                        method=self.feature_selection_method,
                        **get_not_none_kwargs(n_features=self.n_features),
                    ),
                )
            )
        if self.oversampling_method is not None:
            steps.append(
                (
                    "balance",
                    create_oversampling_model(
                        method=self.oversampling_method,
                        random_state=self.random_state,
                    ),
                )
            )
        pipeline = Pipeline(steps)
        return pipeline


def create_oversampling_model(method: str, random_state: int = config.SEED):
    if method is None:
        return None
    if method == "ADASYN":
        return ADASYNWrapper(random_state=random_state)
    elif method == "SMOTE":
        return SMOTEWrapper(random_state=random_state)
    elif method == "BorderlineSMOTE":
        return BorderlineSMOTEWrapper(
            random_state=random_state, kind="borderline1"
        )
    raise ValueError(f"Unknown oversampling method: {method}")


class MinMaxWrapper(MinMaxScaler):
    def fit_transform(self, X, y=None):
        self.fit(X)
        return super().transform(X), y

    def transform(self, X, y=None):
        return super().transform(X), y


class ADASYNWrapper(ADASYN):
    def __init__(self, random_state=config.SEED):
        super().__init__(random_state=random_state)

    def fit_transform(self, data, *args):
        return super().fit_resample(*data)

    def transform(self, X):
        log.debug("ADASYN does nothing on .transform()...")
        return X


class SMOTEWrapper(SMOTE):
    def __init__(self, random_state=config.SEED):
        super().__init__(random_state=random_state)

    def fit_transform(self, data, *args):
        return super().fit_resample(*data)

    def transform(self, X):
        log.debug("SMOTE does nothing on .transform()...")
        return X


class BorderlineSMOTEWrapper(BorderlineSMOTE):
    def __init__(self, kind="borderline-1", random_state=config.SEED):
        super().__init__(kind=kind, random_state=random_state)

    def fit_transform(self, data, *args):
        return super().fit_resample(*data)

    def transform(self, X):
        log.debug("BorderlineSMOTE does nothing on .transform()...")
        return X



import logging
import os
from pathlib import Path
from typing import Optional, Sequence

import pandas as pd
import SimpleITK as sitk
import typer
from nipype.interfaces.dcm2nii import Dcm2niix

from autorad.config.type_definitions import PathLike
from autorad.data.dataset import ImageDataset
from autorad.utils import spatial

log = logging.getLogger(__name__)

dicom_app = typer.Typer()
nrrd_app = typer.Typer()


def generate_border_masks(
    dataset: ImageDataset,
    margin_in_mm: float | Sequence[float],
    output_dir: PathLike,
):
    """
    Generate a border mask (= mask with given margin around the original ROI)
    for each mask in the dataset.
    Returns a DataFrame extending ImageDataset.df with the additional column
    "dilated_mask_path_<margin_in_mm>".
    """
    dilated_paths = []
    for id_, mask_path in zip(dataset.ids, dataset.mask_paths):
        output_path = Path(output_dir) / f"{id_}_border_mask.nii.gz"
        spatial.get_border_outside_mask_mm(
            mask_path, margin_in_mm, output_path
        )
        dilated_paths.append(str(output_path))
    result_df = dataset.df.copy()
    result_df[f"border_mask_path_{margin_in_mm}mm"] = dilated_paths

    return result_df


def get_paths_with_separate_folder_per_case_loose(
    data_dir: PathLike,
    image_stem: str = "image",
    mask_stem: str = "segmentation",
    relative: bool = False,
) -> pd.DataFrame:
    ids, images, masks = [], [], []
    for id_ in os.listdir(data_dir):
        id_dir = os.path.join(data_dir, id_)
        if not os.path.isdir(id_dir):
            continue
        case_paths = Path(id_dir).glob("*.nii.gz")
        image_paths = [p for p in case_paths if image_stem in p.name]
        if not len(image_paths) == 1:
            log.error(
                f"Expected 1, found {len(image_paths)} images for ID={id_}"
            )
            continue
        image_path = image_paths[0]
        for fname in os.listdir(id_dir):
            if mask_stem in fname:
                mask_path = os.path.join(id_dir, fname)
                if relative:
                    image_path = os.path.relpath(image_path, data_dir)
                    mask_path = os.path.relpath(mask_path, data_dir)

                ids.append(id_)
                images.append(str(image_path))
                masks.append(str(mask_path))
    return pd.DataFrame(
        {"case_ID": ids, "image_path": images, "segmentation_path": masks}
    )


def get_paths_with_separate_folder_per_case(
    data_dir: PathLike,
    image_stem: str = "image",
    mask_stem: str = "segmentation",
    relative: bool = False,
) -> pd.DataFrame:
    ids, images, masks = [], [], []
    for id_ in os.listdir(data_dir):
        id_dir = os.path.join(data_dir, id_)
        if not os.path.isdir(id_dir):
            continue
        image_path = os.path.join(id_dir, f"{image_stem}.nii.gz")
        mask_path = os.path.join(id_dir, f"{mask_stem}.nii.gz")
        if not os.path.exists(image_path):
            log.error(f"Image for ID={id_} does not exist ({image_path})")
            continue
        if not os.path.exists(mask_path):
            log.error(f"Mask for ID={id_} does not exist ({mask_path})")
            continue
        if relative:
            image_path = os.path.relpath(image_path, data_dir)
            mask_path = os.path.relpath(mask_path, data_dir)

        ids.append(id_)
        images.append(image_path)
        masks.append(mask_path)
    return pd.DataFrame(
        {"ID": ids, "image_path": images, "segmentation_path": masks}
    )


def get_dcm2niix_converter(
    dicom_dir: Path, save_dir: Path, out_filename: Optional[str] = None
):
    """
    Args:
        dicom_dir: directory with DICOM Study
        save_dir: directory where to save the nifti
        out_filename: what to include in filenames (from dcm2niix)
    Returns:
        converter: nipype Dcm2niix object based on rodenlab dcm2niix with basic
                   default parameters to merge Dicom series into 3d Volumes
    """
    converter = Dcm2niix()
    converter.inputs.source_dir = str(dicom_dir)
    converter.inputs.output_dir = str(save_dir)
    converter.inputs.merge_imgs = True
    if out_filename:
        converter.inputs.out_filename = out_filename

    return converter


@dicom_app.command()
def dicom_to_nifti(
    input_dir: str,
    output_dir: str,
    subdir_name: str = "",
    out_filename: Optional[str] = None,
):
    """
    Args:
        input_dir: absolute path to the directory with all the cases containing
                dicoms
        output_dir: absolute path to the directory where to save nifties
        subdir_name: optional name of subdirectory within case dir
    """
    for id_ in os.listdir(input_dir):
        dicom_dir = Path(input_dir) / id_ / subdir_name
        save_dir = Path(output_dir) / id_ / subdir_name
        save_dir.mkdir(exist_ok=True, parents=True)
        if not dicom_dir.exists():
            raise FileNotFoundError(f"Dicom directory {dicom_dir} not found.")
        converter = get_dcm2niix_converter(dicom_dir, save_dir, out_filename)
        converter.run()


@nrrd_app.command()
def nrrd_to_nifti(nrrd_path: str, output_path: str):
    """
    Converts an image in NRRD format to NifTI format.
    """
    img = sitk.ReadImage(nrrd_path)
    os.makedirs(output_path, exist_ok=True)
    sitk.WriteImage(img, str(output_path))



import os
from pathlib import Path

import mlflow

from autorad.config.type_definitions import PathLike


def get_model_by_name(name, models):
    for model in models:
        if model.name == name:
            return model
    raise ValueError(f"Model with name {name} not found")


def init_mlflow(registry_dir):
    registry_dir = Path(registry_dir)
    registry_dir.mkdir(parents=True, exist_ok=True)
    mlflow.set_tracking_uri("file://" + str(registry_dir.absolute()))


def mlflow_dashboard(experiment_dir: PathLike):
    command = f"mlflow server -h 0.0.0.0 -p 5000 --backend-store-uri file://{str(experiment_dir)} &"
    os.system(command)


def log_mlflow_params(params):
    for param_name, param_value in params.items():
        mlflow.log_param(param_name, param_value)



import json
import logging
import os
import shutil

import numpy as np
import pandas as pd
from monai.transforms import LoadImage

log = logging.getLogger(__name__)


def make_if_dont_exist(folder_path, overwrite=False):
    """
    creates a folder if it does not exists
    input:
    folder_path : relative path of the folder which needs to be created
    over_write :(default: False) if True overwrite the existing folder
    """
    if os.path.exists(folder_path):

        if not overwrite:
            log.info(f"{folder_path} exists.")
        else:
            log.info(f"{folder_path} overwritten.")
            shutil.rmtree(folder_path)
            os.makedirs(folder_path)

    else:
        os.makedirs(folder_path)
        log.info(f"{folder_path} created!")


def load_json(file_name):
    with open(file_name) as f:
        data = json.load(f)
    return data


def save_json(data, output_path):
    with open(output_path, "w") as f:
        json.dump(data, f, indent=4)


def load_image(img_path) -> np.ndarray:
    img = LoadImage()(img_path)
    return img[0]


def save_predictions_to_csv(y_true, y_pred, output_path):
    predictions = pd.DataFrame(
        {"y_true": y_true, "y_pred_proba": y_pred}
    ).sort_values("y_true", ascending=False)
    predictions.to_csv(output_path, index=False)



from pathlib import Path

import pandas as pd
import streamlit as st

from autorad.config import config
from autorad.external.download_WORC import download_WORCDatabase
from autorad.utils import preprocessing


def download_example_dataset():
    if st.button("Download example dataset"):
        dataset_dir = Path(config.RESULT_DIR) / "WORC_desmoid"
        with st.spinner("Downloading dataset..."):
            download_WORCDatabase(
                dataset="Desmoid",
                data_folder=dataset_dir,
                n_subjects=10,
            )
        st.success(f"Downloaded example dataset to {dataset_dir}")


def show():
    with st.sidebar:
        load_example_data = st.checkbox(
            "Download example dataset to the results directory"
        )
    if load_example_data:
        download_example_dataset()

    st.write(
        """
    #####
    Before you proceed with the next tasks you need to create a .csv table.
    The table should contain the following information for each case:
    - **ID** (has to be unique)
    - Path to the **image**
    - Path to the **segmentation**
    - Optionally: **Label** (0 or 1)

    Example:
    """
    )
    example_path_df = pd.read_csv(
        Path(__file__).parents[2] / "paths_example.csv"
    )
    st.write(example_path_df)
    st.write(
        """
    #####
    You can create the table manually (e.g. in Excel) or, if your dataset
    follows the specified structure, you can **generate it below** ⬇
    """
    )
    dir_structures = {
        "Separate folder for each case": """
            Expected format:
            ```
            <dataset folder>
            +-- case1
            |   +-- image.nii.gz
            |   +-- segmentation.nii.gz
            +-- case2
            |   +-- image.nii.gz
            |   +-- segmentation.nii.gz
            ...
            ```
            """,
        "Separate folder for each case with multiple segmentations per image": """
            Expected format:
            ```
            <dataset folder>
            +-- case1
            |   +-- image.nii.gz
            |   +-- segmentation1.nii.gz
            |   +-- segmentation002.nii.gz
            |   +-- segmentation_03.nii.gz
            ...
            +-- case2
            |   +-- image.nii.gz
            |   +-- segmentation_first.nii.gz
            |   +-- segmentation_second.nii.gz
            ...
            ```
            """,
        "One folder for all the images, and one for the segmentations": """
            Expected format:
            ```
            <dataset folder>
            +-- images
            |   +-- 1.nii.gz
            |   +-- 2.nii.gz
            |   +-- ...
            +-- segmentations
            |   +-- 1.nii.gz
            |   +-- 2.nii.gz
            |   +-- ...
            ...
            ```
            """,
    }
    formats = list(dir_structures.keys())[:2]  # no support for 3rd option
    col1, col2 = st.columns(2)
    with col1:
        format = st.radio("Dataset structure:", formats)
    with col2:
        st.write(dir_structures[format])
    data_dir = st.text_input(
        "Enter path to the root directory of your dataset:",
        value=config.INPUT_DIR,
    )
    if not data_dir:
        st.stop()
    data_dir = Path(data_dir).absolute()
    if not data_dir.is_dir():
        st.error(f"The entered path is not a directory ({data_dir})")
    else:
        st.success(f"The entered path is a directory! ({data_dir})")
        col1, col2 = st.columns(2)
        with col1:
            image_stem = st.text_input(
                "How to identify image files? What root does every filename contain?",
                value="image",
            )
        with col2:
            mask_stem = st.text_input(
                "How to identify segmentation files? What root does every filename contain?",
                value="segmentation",
            )
        if format == formats[0]:
            paths_df = preprocessing.get_paths_with_separate_folder_per_case(
                data_dir=data_dir,
                image_stem=image_stem,
                mask_stem=mask_stem,
            )
        else:
            paths_df = (
                preprocessing.get_paths_with_separate_folder_per_case_loose(
                    data_dir=data_dir,
                    image_stem=image_stem,
                    mask_stem=mask_stem,
                )
            )
        if paths_df.empty:
            st.warning(
                "Looks like no files were found."
                " Please check the dataset root directory and file names."
            )
        else:
            st.success(f"Found {len(paths_df)} cases.")
        st.download_button(
            "Download table ⬇️",
            paths_df.to_csv(index=False),
            file_name="paths.csv",
        )
        st.dataframe(paths_df)


if __name__ == "__main__":
    show()



import logging
from multiprocessing import Pool

import pandas as pd
import streamlit as st

from autorad.config.type_definitions import PathLike
from autorad.feature_extraction.extractor import FeatureExtractor
from autorad.utils.utils import time_it

log = logging.getLogger(__name__)


class StreamlitFeatureExtractor(FeatureExtractor):
    def __init__(
        self,
        dataset,
        feature_set="pyradiomics",
        extraction_params="Baessler_CT.yaml",
        n_jobs=None,
    ):
        super().__init__(dataset, feature_set, extraction_params, n_jobs)

    @time_it
    def get_features(self):
        """
        Run extraction for all cases.
        """
        progressbar = st.progress(0)
        lst_of_feature_dicts = []
        for i, (image_path, mask_path) in enumerate(
            zip(self.dataset.image_paths, self.dataset.mask_paths)
        ):
            feature_dict = self.get_features_for_single_case(
                image_path, mask_path
            )
            lst_of_feature_dicts.append(feature_dict)
            fraction_complete = (i + 1) / len(self.dataset.image_paths)
            progressbar.progress(fraction_complete)
        feature_df = pd.DataFrame(lst_of_feature_dicts)
        return feature_df

    def get_features_for_single_case_from_zipped_paths(
        self, paths: tuple[PathLike, PathLike]
    ) -> dict | None:
        image_path, mask_path = paths
        return super().get_features_for_single_case(image_path, mask_path)

    @time_it
    def get_features_parallel(self):
        mask_paths = [path_ for path_ in self.dataset.mask_paths]
        image_paths = [path_ for path_ in self.dataset.image_paths]

        progressbar = st.progress(0)
        n = len(image_paths)
        p = Pool(self.n_jobs)
        lst_of_feature_dicts = []
        for i, result in enumerate(
            p.imap(
                self.get_features_for_single_case_from_zipped_paths,
                zip(image_paths, mask_paths),
            )
        ):
            lst_of_feature_dicts.append(result)
            fraction_complete = (i + 1) / n
            progressbar.progress(fraction_complete)
        feature_df = pd.DataFrame(lst_of_feature_dicts)
        return feature_df



import os
import importlib.util
from pathlib import Path
from shutil import copy
import zipfile
import pandas as pd
import streamlit as st
from ruamel.yaml import YAML
import tempfile
import git


def zip_directory(folder_path, zip_path):
    with zipfile.ZipFile(zip_path, mode="w") as zipf:
        len_dir_path = len(folder_path)
        for root, _, files in os.walk(folder_path):
            for file in files:
                file_path = os.path.join(root, file)
                zipf.write(file_path, file_path[len_dir_path:])


# from https://github1s.com/jrieke/traingenerator
#      /blob/HEAD/app/utils.py#L1-L177
def import_from_file(module_name: str, filepath: str):
    """
    Imports a module from file.

    Args:
        module_name (str): Assigned to the module's __name__ parameter
        (does not influence how the module is named outside of this function)
        filepath (str): Path to the .py file

    Returns:
        The module
    """
    spec = importlib.util.spec_from_file_location(module_name, filepath)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def upload_file(label):
    uploaded_file = st.file_uploader(label)
    if uploaded_file is None:
        st.stop()
    return uploaded_file


def load_df(label):
    """ """
    uploaded_file = upload_file(label)
    df = pd.read_csv(uploaded_file)
    return df


def load_test_data(out_dir):
    """
    Loads test data to the `out_dir` directory
    """
    test_out_dir = Path(out_dir) / "test_data"
    test_out_dir.mkdir(exist_ok=True)
    if not dir_nonempty(test_out_dir):
        repo_address = "https://github.com/pwoznicki/AutoRadiomics.git"
        with tempfile.TemporaryDirectory() as tmp_dir:
            git.Git(tmp_dir).clone(repo_address)
            tmp_dir = Path(tmp_dir)
            for fpath in tmp_dir.rglob("*.nii.gz"):
                copy(fpath, test_out_dir)


def read_yaml(yaml_path):
    """
    Reads .yaml file and returns a dictionary
    """
    yaml_path = Path(yaml_path)
    yaml = YAML(typ="safe")
    data = yaml.load(yaml_path)
    return data


def save_yaml(data, yaml_path):
    """
    Saves a dictionary to .yaml file
    """
    yaml_path = Path(yaml_path)
    yaml = YAML(typ="safe")
    yaml.dump(data, yaml_path)


def dir_nonempty(dir_path):
    path = Path(dir_path)
    if not path.is_dir():
        return False
    has_next = next(path.iterdir(), None)
    if has_next is None:
        return False
    return True


# # from https://github.com/streamlit/streamlit/issues/1019
# def folder_picker(label="Select folder"):
#     root = tk.Tk()
#     root.withdraw()

#     root.wm_attributes("-topmost", 1)

#     # Add button
#     clicked = st.button(label=label)
#     if clicked:
#         dirname = st.text_input(
#             "Selected folder:", filedialog.askdirectory(master=root)
#         )
#     return dirname



import json
import logging
import logging.config
import os
import sys
import tempfile
from pathlib import Path

from rich.logging import RichHandler

import autorad

CONFIG_DIR = os.path.dirname(__file__)
TEST_DATA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(autorad.__file__)),
    "tests",
    "testing_data",
)
if "INPUT_DIR" in os.environ:
    INPUT_DIR = os.environ["INPUT_DIR"]
else:
    INPUT_DIR = tempfile.mkdtemp()
if "RESULT_DIR" in os.environ:
    RESULT_DIR = os.environ["RESULT_DIR"]
else:
    RESULT_DIR = tempfile.mkdtemp()

MODEL_REGISTRY = os.path.join(RESULT_DIR, "models")
# os.makedirs(MODEL_REGISTRY, exist_ok=True)
# mlflow.set_tracking_uri("file://" + MODEL_REGISTRY)

PARAM_DIR = os.path.join(CONFIG_DIR, "pyradiomics_params")
PRESETS = {
    "CT default": "default_feature_map.yaml",
    "CT reproducibility (Baessler et al.)": "Baessler_CT.yaml",
}
with open(os.path.join(CONFIG_DIR, "pyradiomics_feature_names.json")) as f:
    PYRADIOMICS_FEATURE_NAMES = json.load(f)

AVAILABLE_CLASSIFIERS = [
    "Random Forest",
    "Logistic Regression",
    "SVM",
    "XGBoost",
]
FEATURE_SELECTION_METHODS = ["anova", "lasso", "boruta", "boruta-shap"]
OVERSAMPLING_METHODS = ["SMOTE", "ADASYN", None]

SEED = 123

MONAI_DATA_DIR = tempfile.mkdtemp()

IS_DEMO = False

# Logging
LOGS_DIR = Path(RESULT_DIR, "logs")
LOGS_DIR.mkdir(parents=True, exist_ok=True)

logging_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "minimal": {"format": "%(message)s"},
        "detailed": {
            "format": "%(levelname)s %(asctime)s [%(name)s:%(filename)s:%(funcName)s:%(lineno)d]\n%(message)s\n"
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "stream": sys.stdout,
            "formatter": "minimal",
            "level": logging.DEBUG,
        },
        "info": {
            "class": "logging.handlers.RotatingFileHandler",
            "filename": Path(LOGS_DIR, "info.log"),
            "maxBytes": 10485760,  # 1 MB
            "backupCount": 10,
            "formatter": "detailed",
            "level": logging.INFO,
        },
        "error": {
            "class": "logging.handlers.RotatingFileHandler",
            "filename": Path(LOGS_DIR, "error.log"),
            "maxBytes": 10485760,  # 1 MB
            "backupCount": 10,
            "formatter": "detailed",
            "level": logging.ERROR,
        },
    },
    "root": {
        "handlers": ["console", "info", "error"],
        "level": logging.INFO,
        "propagate": True,
    },
}

logging.config.dictConfig(logging_config)
logger = logging.getLogger()
logger.handlers[0] = RichHandler(markup=True)



import logging
from functools import wraps
from typing import Callable, Union

import numpy as np
import pandas as pd
from scipy import stats
from sklearn.metrics import roc_auc_score
from sklearn.utils import resample
from statsmodels.stats.contingency_tables import mcnemar

log = logging.getLogger(__name__)


def round_up_p(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        p = f(*args, **kwargs)
        rounded_p = np.round(p, 3)
        return rounded_p

    return wrapper


@round_up_p
def compare_groups_not_normally_distributed(
    x: list[float], y: list[float], alternative="two-sided"
):
    """
    Mann-Whitney test (= unpaired Wilcoxon test).
    """
    _, p = stats.ranksums(x, y, alternative=alternative)
    return p


@round_up_p
def compare_age_between_groups(x: list[float], y: list[float]) -> float:
    """
    Perform Welsh's t-test (good when cohorts differ in size,
    because doesn't assume equal variance).
    """
    if not x or not y:
        raise ValueError("x and y must be non-empty lists of strings")
    if any(elem < 0 for elem in (x + y)):
        raise ValueError("Age cannot be negative.")
    _, p = stats.ttest_ind(x, y, equal_var=False)
    return p


@round_up_p
def compare_gender_between_groups(
    genders: list[Union[int, str]], groups: list[Union[int, str]]
) -> int:
    """
    Performs Chi square test for independence.
    Tests if observed frequencies are independent of the expected
    frequencies.
    To be used for categorical variables, e.g. the gender distributions.

    >>> genders = ['m', 'f', 'm', 'f']
    >>> groups = ['train', 'train', 'test', 'test']
    >>> compare_gender_between_groups(genders, groups)
    1.0

    >>> genders = [1, 1, 1, 0, 0, 0]
    >>> groups = [0, 0, 0, 1, 1, 1]
    >>> compare_gender_between_groups(genders, groups)
    0.102
    """
    contingency_matrix = pd.crosstab(index=genders, columns=groups)
    _, p, _, _ = stats.chi2_contingency(contingency_matrix)
    return p


def compare_sensitivity_mcnemar(y_pred_proba_1, y_pred_proba_2):
    """
    Compare sensitivity of two models using McNemar's test
    """
    contingency_table = pd.crosstab(
        index=y_pred_proba_1, columns=y_pred_proba_2
    )
    _, p = mcnemar(contingency_table)
    return p


def bootstrap_auc(y_true, y_pred_proba):
    """
    Get AUC and 95% Confidence Interval from bootstrapping.
    """
    sample_statistic, lower, upper = bootstrap_statistic(
        roc_auc_score,
        y_true,
        y_pred_proba,
    )

    return sample_statistic, lower, upper


def bootstrap_statistic(statistic: Callable, x, y, num_folds=1000):
    """
    Bootstrap statistic for comparing two groups.
    Args:
        statistic: function that takes two lists of values and returns a
            statistic.
        x: list of values for group 1
        y: list of values for group 2
        num_folds: number of bootstrap samples to draw
    Returns:
        statistic: sample statistic for the two groups
        lower_bound: lower bound of the 95% confidence interval
        upper_bound: upper bound of the 95% confidence interval
    """
    stats = []
    for i in range(num_folds):
        boot_x, boot_y = resample(
            x, y, replace=True, n_samples=len(x), random_state=i
        )
        stat = statistic(boot_x, boot_y)
        stats.append(stat)
    stats_arr = np.array(stats)
    sample_statistic = statistic(x, y)
    lower_bound = np.percentile(stats_arr, 2.5)
    upper_bound = np.percentile(stats_arr, 97.5)

    return sample_statistic, lower_bound, upper_bound



import streamlit as st
import template_utils

from autorad.config import config


def show():
    # model = mlflow.pyfunc.load_model(model_path)
    col1, col2 = st.columns(2)
    with col1:
        img_path = st.file_uploader(label="Input image")
    with col2:
        mask_path = st.file_uploader(label="Input segmentation")
    # model.predict()
    save_dir = config.RESULT_DIR
    if st.button("Predict"):
        st.write("Predicting...")
        template_utils.extract_feature_maps(img_path, mask_path, save_dir)


if __name__ == "__main__":
    show()



from typing import Sequence, Union

import numpy as np
from sklearn.model_selection import StratifiedKFold, train_test_split

from autorad.config import config


def split_cross_validation(
    ids: Sequence[Union[str, int]],
    labels: Sequence[Union[str, int]],
    n_splits: int = 5,
    random_state: int = config.SEED,
) -> dict:
    """
    Split data into n_splits folds for cross-validation, with stratification.
    """
    ids_array = np.asarray(ids)
    ids_split = {}
    cv = StratifiedKFold(
        n_splits=n_splits, shuffle=True, random_state=random_state
    )
    for i, (train_indices, test_indices) in enumerate(cv.split(ids, labels)):
        train_ids = ids_array[train_indices].tolist()
        test_ids = ids_array[test_indices].tolist()
        ids_split[f"fold_{i}"] = (train_ids, test_ids)
    return ids_split


def split_full_dataset(
    ids: Sequence[str],
    labels: Sequence[str],
    test_size: float = 0.2,
    n_splits: int = 5,
    random_state: int = config.SEED,
) -> dict:
    """
    Split data into test and training, then divide training into n folds for
    cross-validation. Labels are needed for stratification.
    Save the splits as json.
    """
    ids_train, ids_test, y_train, y_test = train_test_split(
        ids,
        labels,
        test_size=test_size,
        stratify=labels,
        random_state=random_state,
    )
    ids_train_cv = split_cross_validation(
        ids_train, y_train, n_splits, random_state=random_state
    )
    ids_split = {
        "split_type": f"{test_size:.0%} test + {n_splits}-fold "
        "cross-validation on the training set",
        "test": ids_test,
        "train": ids_train_cv,
    }
    return ids_split


def split_train_val_test(
    ids: Sequence[str],
    labels: Sequence[str],
    test_size: float = 0.2,
    val_size: float = 0.2,
    random_state: int = config.SEED,
) -> dict:
    """
    Stratified train/val/test split without cross-validation.
    """
    ids_train_val, ids_test, y_train_val, y_test = train_test_split(
        ids,
        labels,
        test_size=test_size,
        stratify=labels,
        random_state=random_state,
    )

    val_proportion_without_test = val_size / (1 - test_size)
    ids_train, ids_val, y_train, y_val = train_test_split(
        ids_train_val,
        y_train_val,
        test_size=val_proportion_without_test,
        stratify=y_train_val,
        random_state=random_state,
    )
    train_size = 1 - val_size - test_size
    ids_split = {
        "split_type": f"stratified split: {train_size:.0%} train + {val_size:.0%} validation"
        f" + {test_size:.0%} test",
        "train": ids_train,
        "val": ids_val,
        "test": ids_test,
    }
    return ids_split



from __future__ import annotations

import abc
import logging
import warnings
from typing import Sequence

import numpy as np
import pandas as pd
from boruta import BorutaPy

# from BorutaShap import BorutaShap
from sklearn.ensemble import RandomForestClassifier
from sklearn.exceptions import ConvergenceWarning
from sklearn.feature_selection import SelectKBest, f_classif
from sklearn.linear_model import Lasso
from sklearn.model_selection import GridSearchCV

from autorad.config import config

warnings.filterwarnings(action="ignore", category=ConvergenceWarning)
log = logging.getLogger(__name__)


class NoFeaturesSelectedError(Exception):
    """raised when feature selection fails"""

    pass


class CoreSelector(abc.ABC):
    """Template for feature selection methods"""

    def __init__(self):
        self.selected_columns = None

    @abc.abstractmethod
    def fit(self, X: np.ndarray, y: np.ndarray) -> list[int]:
        """fit method should update self.selected_columns.
        If no features are selected, it should raise
        NoFeaturesSelectedError.
        """
        pass

    def fit_transform(
        self, X: np.ndarray, y: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        self.fit(X, y)
        return X[:, self.selected_columns], y

    def transform(self, X: np.ndarray) -> np.ndarray:
        if self.selected_columns is None:
            raise NoFeaturesSelectedError(
                "No features selected!" "Call fit() first before transforming."
            )
        return X[:, self.selected_columns]

    def selected_features(self, column_names: Sequence[str]):
        try:
            selected_features = [
                column_names[i] for i in self.selected_columns
            ]
        except NoFeaturesSelectedError as e:
            raise e

        return selected_features


class AnovaSelector(CoreSelector):
    def __init__(self, n_features: int = 10):
        self.n_features = n_features
        self.model = SelectKBest(f_classif, k=self.n_features)
        super().__init__()

    def fit(self, X, y):
        self.model.fit(X, y)
        support = self.model.get_support(indices=True)
        if support is None:
            raise NoFeaturesSelectedError("ANOVA failed to select features.")
        self.selected_columns = support.tolist()


class LassoSelector(CoreSelector):
    def __init__(self, alpha=0.002):
        self.model = Lasso(random_state=config.SEED, alpha=alpha)
        super().__init__()

    def optimize_params(self, X, y, verbose=0):
        search = GridSearchCV(
            self.model,
            {"alpha": np.logspace(-5, 1, num=100)},
            cv=5,
            scoring="neg_mean_squared_error",
            verbose=verbose,
        )
        search.fit(X, y)
        best_params = search.best_params_
        log.info(f"Best params for Lasso: {best_params}")
        self.model = self.model.set_params(**best_params)

    def fit(self, X, y):
        self.model.fit(X, y)
        coefficients = self.model.coef_
        importance = np.abs(coefficients)
        self.selected_columns = np.where(importance > 0)[0].tolist()
        if not self.selected_columns:
            raise NoFeaturesSelectedError("Lasso failed to select features.")

    def params_to_optimize(self):
        return {"alpha": np.logspace(-5, 1, num=100)}


class BorutaSelector(CoreSelector):
    def fit(self, X, y, verbose=0):
        model = BorutaPy(
            RandomForestClassifier(
                max_depth=5, n_jobs=-1, random_state=config.SEED
            ),
            n_estimators="auto",
            verbose=verbose,
            random_state=config.SEED,
        )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model.fit(X, y)
        self.selected_columns = np.where(model.support_)[0].tolist()
        if not self.selected_columns:
            raise NoFeaturesSelectedError("Boruta failed to select features.")

    # class BorutaSHAPSelector(CoreSelector):
    #     def fit(self, X, y, verbose=0):
    #         model = BorutaShap(importance_measure="shap", classification=True)
    #         # BorutaShap requires X to be pd.DataFrame
    #         colnames = np.arange(X.shape[1]).astype(str)
    #         X_df = pd.DataFrame(X, columns=colnames)
    #         model.fit(
    #             X=X_df, y=y, n_trials=100, sample=False, verbose=bool(verbose)
    #         )
    #         selected_columns_str = model.Subset().columns
    #         self.selected_columns = [int(c) for c in selected_columns_str]


class FeatureSelectorFactory:
    def __init__(self):
        self.selectors = {
            "anova": AnovaSelector,
            "lasso": LassoSelector,
            "boruta": BorutaSelector,
            # "boruta-shap": BorutaSHAPSelector,
        }

    def register_selector(self, name, selector):
        self.selectors[name] = selector

    def get_selector(self, name, *args, **kwargs):
        selector = self.selectors[name]
        if not selector:
            raise ValueError(f"Unknown feature selection ({name}).")
        return selector(*args, **kwargs)


def create_feature_selector(
    method: str = "anova",
    *args,
    **kwargs,
):
    selector = FeatureSelectorFactory().get_selector(method, *args, **kwargs)
    return selector



import base64
import logging
import math
import os
import re
import uuid
from pathlib import Path

import jupytext
import radiomics
import SimpleITK as sitk
import streamlit as st
from radiomics import featureextractor

from autorad.config import config
from autorad.data.dataset import ImageDataset
from autorad.webapp import utils


def guess_idx_of_img_colname(colnames):
    for i, colname in enumerate(colnames):
        if "img" in colname or "image" in colname:
            return i
    return 0


def guess_idx_of_seg_colname(colnames):
    for i, colname in enumerate(colnames):
        if "seg" in colname or "mask" in colname:
            return i
    return 0


def load_path_df():
    path_df = utils.load_df("Choose a CSV file with paths:")
    st.dataframe(path_df)
    col1, col2, col3 = st.columns(3)
    colnames = path_df.columns.tolist()
    with col1:
        image_col = st.selectbox(
            "Path to image",
            colnames,
            index=guess_idx_of_img_colname(colnames),
            key=uuid.uuid4(),
        )
    with col2:
        mask_col = st.selectbox(
            "Path to segmentation",
            colnames,
            index=guess_idx_of_seg_colname(colnames),
            key=uuid.uuid4(),
        )
    with col3:
        id_col = st.selectbox(
            "ID (optional)", [None] + colnames, key=uuid.uuid4()
        )
    path_df.dropna(subset=[image_col, mask_col], inplace=True)
    dataset = ImageDataset(
        df=path_df,
        image_colname=image_col,
        mask_colname=mask_col,
        ID_colname=id_col,
        root_dir=config.INPUT_DIR,
    )
    return dataset


def extract_feature_maps(image_path, seg_path, save_dir):
    extraction_params = radiomics_params_voxelbased()
    radiomics.setVerbosity(logging.INFO)
    extractor = featureextractor.RadiomicsFeatureExtractor(extraction_params)
    feature_vector = extractor.execute(image_path, seg_path, voxelBased=True)
    for feature_name, feature_value in feature_vector.items():
        if isinstance(feature_value, sitk.Image):
            save_path = save_dir / f"{feature_name}.nii.gz"
            save_path = str(save_path)
            sitk.WriteImage(feature_value, save_path)


def radiomics_params():
    param_dir = Path(config.PARAM_DIR)
    presets = config.PRESETS
    preset_options = list(presets.keys())
    name = st.selectbox(
        "Choose a preset with parameters for feature extraction",
        preset_options,
        key=uuid.uuid4(),
    )
    preset_setup = utils.read_yaml(param_dir / presets[name])
    final_setup = preset_setup.copy()

    with st.expander("Manually edit the extraction parameters"):
        final_setup = select_classes(preset_setup, final_setup)
        setting = preset_setup["setting"]
        st.write(""" Filters: """)
        filter = preset_setup["imageType"]
        col1, col2, col3 = st.columns(3)
        with col1:
            st.checkbox(
                "original", value=("Original" in filter), key=uuid.uuid4()
            )
        with col2:
            turn_on_log = st.checkbox(
                "Laplacian of Gaussian",
                value=("LoG" in filter),
                key=uuid.uuid4(),
            )
            if turn_on_log:
                sigmas = filter["LoG"]["sigma"]
                for sigma in sigmas:
                    st.number_input("sigma", value=sigma, key=uuid.uuid4())
        with col3:
            st.checkbox(
                "Wavelet", value=("Wavelet" in filter), key=uuid.uuid4()
            )
        col1, col2, col3 = st.columns(3)
        with col1:
            st.write(""" ##### Normalization: """)
            normalization = setting["normalize"]
            is_normalized = st.checkbox(
                "Normalize", value=normalization, key=uuid.uuid4()
            )
            if is_normalized:
                final_setup["setting"]["normalize"] = True
            else:
                final_setup["setting"]["normalize"] = False
        with col2:
            bin_width = st.number_input(
                """Bin width:""", value=setting["binWidth"], key=uuid.uuid4()
            )
            final_setup["setting"]["binWidth"] = bin_width
        with col3:
            label = st.number_input(
                "Label:", value=setting["label"], key=uuid.uuid4()
            )
            final_setup["setting"]["label"] = int(label)
        st.write(""" #### Full parameter file: """, preset_setup)
    return preset_setup


def choose_preset():
    config_dir = os.path.dirname(config.__file__)
    param_dir = Path(config_dir) / "pyradiomics_params"
    presets = config.PRESETS
    preset_options = list(presets.keys())
    name = st.selectbox(
        "Choose a preset with parameters for feature extraction",
        preset_options,
        key=uuid.uuid4(),
    )
    preset_setup = utils.read_yaml(param_dir / presets[name])
    return preset_setup


def update_setup_with_class(setup, class_name, class_active):
    assert "featureClass" in setup
    if class_active[class_name]:
        setup["featureClass"][class_name] = []
    else:
        setup["featureClass"].pop(class_name, None)
    return setup


def select_classes(preset_setup, final_setup, exclude_shape=False):
    st.write(""" #### Select classes: """)

    all_feature_names = config.PYRADIOMICS_FEATURE_NAMES
    all_classes = list(all_feature_names.keys())
    if exclude_shape:
        if "shape" in all_classes:
            all_classes.remove("shape")
    preset_classes = preset_setup["featureClass"]
    class_active = {}
    cols = st.columns(len(all_classes))
    for i, class_name in enumerate(all_classes):
        with cols[i]:
            class_active[class_name] = st.checkbox(
                class_name,
                value=(class_name in preset_classes),
                key=uuid.uuid4(),
            )
        final_setup = update_setup_with_class(
            final_setup, class_name, class_active
        )
    # st.write("##### Optional: select specific features within class:")
    st.info(
        """
        If you select a class, all its features are included by default.
        You can however select only specific features (optional):
        """
    )
    cols = st.columns(len(preset_classes.keys()))
    for i, class_name in enumerate(preset_classes.keys()):
        with cols[i]:
            if class_active[class_name]:
                feature_names = st.multiselect(
                    class_name, all_feature_names[class_name], key=uuid.uuid4()
                )
                final_setup["featureClass"][class_name] = feature_names
    return final_setup


def radiomics_params_voxelbased():
    param_dir = Path(config.PARAM_DIR)
    presets = config.PRESETS
    preset_options = list(presets.keys())
    name = st.selectbox(
        "Choose a preset with parameters for feature extraction",
        preset_options,
        key=uuid.uuid4(),
    )
    preset_setup = utils.read_yaml(param_dir / presets[name])
    if "shape" in preset_setup["featureClass"]:
        preset_setup["featureClass"].pop("shape", None)
    final_setup = preset_setup.copy()

    with st.expander("Manually edit the extraction parameters"):
        final_setup = select_classes(
            preset_setup, final_setup, exclude_shape=True
        )

        setting = preset_setup["setting"]
        col1, col2, col3 = st.columns(3)
        with col1:
            st.write(""" ##### Normalization: """)
            normalization = setting["normalize"]
            is_normalized = st.checkbox(
                "Normalize", value=normalization, key=uuid.uuid4()
            )
            if is_normalized:
                final_setup["setting"]["normalize"] = True
            else:
                final_setup["setting"]["normalize"] = False
        with col2:
            bin_width = st.number_input(
                """Bin width:""", value=setting["binWidth"], key=uuid.uuid4()
            )
            final_setup["setting"]["binWidth"] = bin_width
        with col3:
            label = st.number_input(
                "Label:", value=setting["label"], key=uuid.uuid4()
            )
            final_setup["setting"]["label"] = int(label)
        st.write(""" #### Full parameter file: """, preset_setup)

    return final_setup


def code_header(text):
    """
    Insert section header into a jinja file, formatted as Python comment.
    Leave 2 blank lines before the header.
    """
    seperator_len = (75 - len(text)) / 2
    seperator_len_left = math.floor(seperator_len)
    seperator_len_right = math.ceil(seperator_len)
    return f"# {'-' * seperator_len_left} {text} {'-' * seperator_len_right}"


def notebook_header(text):
    """
    Insert section header into a jinja file, formatted as notebook cell.
    Leave 2 blank lines before the header.
    """
    return f"""# # {text}
"""


def to_notebook(code):
    """Converts Python code to Jupyter notebook format."""
    notebook = jupytext.reads(code, fmt="py")
    return jupytext.writes(notebook, fmt="ipynb")


# borrowed from
# https://github.com/jrieke/traingenerator/blob/main/app/utils.py
def download_button(
    object_to_download, download_filename, button_text  # , pickle_it=False
):
    """
    Generates a link to download the given object_to_download.
    """
    try:
        # some strings <-> bytes conversions necessary here
        b64 = base64.b64encode(object_to_download.encode()).decode()
    except AttributeError:
        b64 = base64.b64encode(object_to_download).decode()

    button_uuid = str(uuid.uuid4()).replace("-", "")
    button_id = re.sub(r"\d+", "", button_uuid)

    custom_css = f"""
        <style>
            #{button_id} {{
                display: inline-flex;
                align-items: center;
                justify-content: center;
                background-color: rgb(255, 255, 255);
                color: rgb(38, 39, 48);
                padding: .25rem .75rem;
                position: relative;
                text-decoration: none;
                border-radius: 4px;
                border-width: 1px;
                border-style: solid;
                border-color: rgb(230, 234, 241);
                border-image: initial;
            }}
            #{button_id}:hover {{
                border-color: rgb(246, 51, 102);
                color: rgb(246, 51, 102);
            }}
            #{button_id}:active {{
                box-shadow: none;
                background-color: rgb(246, 51, 102);
                color: white;
                }}
        </style> """

    dl_link = (
        custom_css
        + f'<a download="{download_filename}" id="{button_id}" href="data:file/txt;base64,{b64}">{button_text}</a><br><br>'
    )

    st.markdown(dl_link, unsafe_allow_html=True)



import json
import os
import sys

import streamlit.bootstrap
import streamlit.config


def _patch_streamlit_print_url():
    _original_print_url = streamlit.bootstrap._print_url

    def _new_print_url(is_running_hello: bool) -> None:
        port = int(streamlit.config.get_option("browser.serverPort"))

        sys.stdout.flush()
        print(json.dumps({"port": port}))
        sys.stdout.flush()

        _original_print_url(is_running_hello)

    streamlit.bootstrap._print_url = _new_print_url


if __name__ == "__main__":
    fname = os.path.join(os.path.dirname(__file__), "webapp/app.py")
    _patch_streamlit_print_url()

    conf = {
        "server.fileWatcherType": "none",
        "server.headless": True,
    }
    streamlit.bootstrap.load_config_options(flag_options=conf)

    streamlit.bootstrap.run(fname, "", args=[], flag_options={})



import os
import webbrowser
from pathlib import Path

import seaborn as sns
import streamlit as st
import utils

from autorad.config import config
from autorad.data.dataset import FeatureDataset
from autorad.models.classifier import MLClassifier
from autorad.training.trainer import Trainer


def show():
    with st.sidebar:
        st.write(
            """
            Given a table with radiomics features extracted in previous step,
            run the training consisting of:
            - Feature selection
            - Hyperparameter optimization for selected models
        """
        )
    feature_df = utils.load_df("Choose a CSV file with radiomics features")
    cm = sns.light_palette("green", as_cmap=True)
    st.dataframe(feature_df.style.background_gradient(cmap=cm))
    all_colnames = feature_df.columns.tolist()
    with st.form("Training config"):
        col1, col2 = st.columns(2)
        with col1:
            target = st.selectbox("Select the label", all_colnames)
        with col2:
            ID_colname = st.selectbox("Select patient ID", all_colnames)
        feature_dataset = FeatureDataset(
            dataframe=feature_df,
            target=target,
            ID_colname=ID_colname,
        )
        col1, col2 = st.columns(2)
        with col1:
            split_method = st.radio(
                label="Splitting",
                options=["train_val_test", "train_with_cross_validation_test"],
            )
        with col2:
            test_size_percent = st.slider(
                label="Test size (%)",
                min_value=0,
                max_value=100,
                value=20,
            )
        test_size = test_size_percent / 100
        model_names = st.multiselect(
            "Select the models", config.AVAILABLE_CLASSIFIERS
        )

        start_training = st.form_submit_button("Start training")
        if start_training:
            run_training_mlflow(
                feature_dataset, split_method, test_size, model_names
            )
    if st.button("Track the models with MLflow dashboard"):
        mlflow_model_dir = str(Path(config.RESULT_DIR) / "models")
        webbrowser.open("http://localhost:8000/")
        os.system(
            f"mlflow server -h 0.0.0.0 -p 8000 --backend-store-uri {mlflow_model_dir}"
        )


def run_training_mlflow(feature_dataset, split_method, test_size, model_names):
    result_dir = config.RESULT_DIR

    feature_dataset.split(
        method=split_method,
        test_size=test_size,
        save_path=(Path(result_dir) / "splits.json"),
    )

    models = [MLClassifier.from_sklearn(name) for name in model_names]
    trainer = Trainer(
        dataset=feature_dataset,
        models=models,
        result_dir=result_dir,
    )
    with st.spinner("Preprocessing in progress..."):
        trainer.run_auto_preprocessing(
            selection_methods=["boruta"], oversampling=False
        )
    trainer.set_optimizer("optuna", n_trials=30)
    with st.spinner("Training in progress..."):
        trainer.run(auto_preprocess=True)
    st.success(
        f"Training done! Predictions saved in your result directory \
        ({result_dir})"
    )


if __name__ == "__main__":
    show()



from pathlib import Path

import streamlit as st

from autorad.config import config
from autorad.utils import preprocessing, testing
from autorad.webapp import template_utils


def show():
    st.write(
        "Once you've created the table from previous step, run some basic tests on the images and masks:"
    )
    dataset = template_utils.load_path_df()
    n_cases = len(dataset.df)
    st.write("##### Test your data:")
    if st.button("Test if ROI has at least one pixel"):
        try:
            testing.check_assertion_dataset(
                testing.assert_has_nonzero, dataset.mask_paths
            )
        except AssertionError as e:
            st.error(e)
        else:
            st.success(f"Passed for all {n_cases} cases!")
    if st.button("Test if image and masks are of the same size"):
        try:
            testing.check_assertion_dataset(
                testing.assert_equal_shape,
                list(zip(dataset.image_paths, dataset.mask_paths)),
            )
        except AssertionError as e:
            st.error(e)
        else:
            st.success(f"Passed for {n_cases} cases!")
    if st.button("Test if ROI in the image has >1 unique values"):
        try:
            testing.check_assertion_dataset(
                testing.assert_has_nonzero_within_roi,
                list(zip(dataset.image_paths, dataset.mask_paths)),
            )
        except AssertionError as e:
            st.error(e)
        else:
            st.success(f"Passed for all {n_cases} cases!")
    st.write(
        """
    #####
    You can proceed to the next step, [**Feature extraction**](#feature-extraction). \n
    Or, if you want to add additional **peritumoral masks** of the region surrounding the segmentation, you can do it below ⬇
    """,
        unsafe_allow_html=True,
    )
    margin = st.number_input(label="Margin (in mm)", value=5)
    margin = int(margin)
    result_dir = Path(config.RESULT_DIR)
    if st.button("Generate border masks"):
        result_dir.mkdir(exist_ok=True)
        dilated_mask_dir = result_dir / f"border_masks_{margin}"
        dilated_mask_dir.mkdir(exist_ok=True, parents=True)
        extended_paths_df = preprocessing.generate_border_masks(
            dataset=dataset,
            margin_in_mm=int(margin),
            output_dir=dilated_mask_dir,
        )
        st.download_button(
            label="Download updated table ⬇️",
            data=extended_paths_df.to_csv(index=False),
            file_name="paths_with_bordermasks.csv",
        )
        st.dataframe(extended_paths_df)


if __name__ == "__main__":
    show()



import seaborn as sns
import streamlit as st
import utils

from autorad.evaluation.evaluator import SimpleEvaluator


def show():
    """Shows the sidebar components for the template and returns
    user inputs as dict."""

    with st.sidebar:
        pass

    # LAYING OUT THE TOP SECTION OF THE APP
    col1, col2 = st.columns((2, 3))

    st.write(
        """
        #####
        Analyze the results with most widely used metrics such as
        AUC ROC curve, precision-recall curve and confusion matrix.
        """
    )
    result_df = utils.load_df("Choose a CSV file with predictions:")
    col1, col2 = st.columns(2)
    with col1:
        cm = sns.light_palette("blue", as_cmap=True)
        st.dataframe(result_df.style.background_gradient(cmap=cm))
    result_df_colnames = result_df.columns.tolist()
    with col2:
        label = st.selectbox("Select the label column", result_df_colnames)
    with col2:
        pred = st.selectbox("Select the prediction column", result_df_colnames)
    # Evaluation
    evaluate = st.button("Evaluate!")
    if evaluate:
        evaluator = SimpleEvaluator(
            y_true=result_df[label].tolist(),
            y_pred_proba=result_df[pred].tolist(),
        )
        col1, col2 = st.columns(2)
        with col1:
            st.write(evaluator.plot_roc_curve())
        with col2:
            st.write(evaluator.plot_precision_recall_curve())
        fig = evaluator.plot_waterfall()
        st.plotly_chart(fig, use_container_width=True)


if __name__ == "__main__":
    show()



import logging
from pathlib import Path

import mlflow
import optuna
from optuna.samplers import TPESampler
from sklearn.model_selection import GridSearchCV

from autorad.config import config
from autorad.utils import io

log = logging.getLogger(__name__)


class OptunaOptimizer:
    def __init__(
        self,
        n_trials: int = 30,
    ):
        self.n_trials = n_trials

    def create_study(self, study_name):
        return optuna.create_study(
            direction="maximize",
            study_name=study_name,
            sampler=TPESampler(seed=config.SEED),
        )


class GridSearchOptimizer:
    def __init__(self, dataset, model, param_dir):
        self.dataset = dataset
        self.model = model
        self.param_dir = Path(param_dir)
        self.param_grid = None

    def get_grid_RandomForest(self):
        n_estimators = [200, 600, 1000]
        max_features = ["auto", "sqrt"]
        max_depth = [10, 50, None]
        min_samples_split = [2, 5, 10]
        min_samples_leaf = [1, 4]
        bootstrap = [True, False]

        self.param_grid = {
            "n_estimators": n_estimators,
            "max_features": max_features,
            "max_depth": max_depth,
            "min_samples_split": min_samples_split,
            "min_samples_leaf": min_samples_leaf,
            "bootstrap": bootstrap,
        }
        return self

    def get_grid_XGBoost(self):
        self.param_grid = {
            "learning_rate": [0.05, 0.10, 0.20, 0.30],
            "max_depth": [2, 4, 8, 12, 15],
            "min_child_weight": [1, 3, 7],
            "gamma": [0.0, 0.1, 0.3],
            "colsample_bytree": [0.3, 0.5, 0.7],
        }
        return self

    def get_grid_LogReg(self):
        self.param_grid = {
            "C": [0.001, 0.01, 0.1, 1, 10, 100, 1000],
            "penalty": ["l2", "none"],
        }
        return self

    def get_grid_SVM(self):
        cs = [0.001, 0.01, 0.1, 1, 5, 10]
        gammas = [0.001, 0.01, 0.1, 1]
        kernels = ["rbf"]
        self.param_grid = {"kernel": kernels, "C": cs, "gamma": gammas}
        return self

    def save_params(self, params):
        self.param_dir.mkdir(exist_ok=True)
        save_path = Path(self.param_dir) / (self.model.name + ".json")
        log.info(f"Saving parameters to: {str(save_path)}")
        io.save_json(params, save_path)

    def load_params(self):
        param_path = self.param_dir / (self.model.name + ".json")
        log.info(f"Loading parameters from: {param_path}")
        if param_path.exists():
            optimal_params = io.load_json(param_path)
            self.model.set_params(optimal_params)
        else:
            raise FileNotFoundError(
                "No param file found. Run `tune_hyperparameters` first."
            )

        return self

    def update_model_params_grid_search(self):
        if self.param_grid is None:
            raise ValueError("First select param grid!")
        else:
            param_searcher = GridSearchCV(
                estimator=self.model.classifier,
                param_grid=self.param_grid,
                scoring="roc_auc",
                cv=self.dataset.cv_splits,
                verbose=0,
                n_jobs=-1,
            )
            rs = param_searcher.fit(
                X=self.dataset.X_train, y=self.dataset.y_train
            )
            optimal_params = rs.best_params_
            log.info(
                f"Best params for {self.model.name}: \
                {optimal_params}"
            )
            self.model.set_params(optimal_params)
            self.save_params(optimal_params)
            mlflow.log_param("optimal_params", optimal_params)

            return self

    def get_param_grid(self):
        model_name = self.model.name
        if model_name == "Random Forest":
            self.get_grid_RandomForest()
        elif model_name == "XGBoost":
            self.get_grid_XGBoost()
        elif model_name == "Logistic Regression":
            self.get_grid_LogReg()
        elif model_name == "SVM":
            self.get_grid_SVM()
        else:
            return ValueError(
                f"Hyperparameter tuning for {model_name} not implemented!"
            )

        return self

    def tune_hyperparameters(self):
        try:
            self.get_param_grid()
            self.update_model_params_grid_search()
        except Exception:
            pass

        return self

    def load_or_tune_hyperparameters(self):
        try:
            self.load_params()
        except Exception:
            log.info(
                "Params couldn't be loaded. Starting hyperparameter tuning..."
            )
            self.tune_hyperparameters()

        return self.model



# Adapted from https://github.com/MStarmans91/WORCDatabase/blob/
# development/datadownloader.py

import logging
import os
import shutil
from glob import glob

# Original license:
# Copyright 2016-2021 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandas as pd
import xnat

logging.getLogger("xnat").setLevel(logging.WARNING)

valid_datasets = ["Lipo", "Desmoid", "GIST", "Liver", "CRLM", "Melanoma"]


def download_subject(project, subject, datafolder, session, verbose=False):
    """Download data of a single XNAT subject."""
    download_counter = 0
    resource_labels = list()
    for e in subject.experiments:
        resmap = {}
        experiment = subject.experiments[e]

        for s in experiment.scans:
            scan = experiment.scans[s]
            print(
                ("\tDownloading patient {}, experiment {}, scan {}.").format(
                    subject.label, experiment.label, scan.id
                )
            )
            for res in scan.resources:
                resource_label = scan.resources[res].label
                if resource_label == "NIFTI":
                    # Create output directory
                    outdir = os.path.join(datafolder, f"{subject.label}")
                    os.makedirs(outdir, exist_ok=True)

                    resmap[resource_label] = scan
                    # print(f'resource is {resource_label}')
                    scan.resources[res].download_dir(outdir)
                    resource_labels.append(resource_label)
                    download_counter += 1

    # Parse resources and throw warnings if they not meet the requirements
    subject_name = subject.label
    if download_counter == 0:
        print(
            f"\t[WARNING] Skipping subject {subject_name}: no (suitable) resources found."
        )
        return False

    if "NIFTI" not in resource_labels:
        print(
            f"\t[WARNING] Skipping subject {subject_name}: no NIFTI resources found."
        )
        return False

    # Reorder files to a easier to read structure
    NIFTI_files = glob(
        os.path.join(
            outdir,
            "*",
            "scans",
            "*",
            "resources",
            "NIFTI",
            "files",
            "*.nii.gz",
        )
    )
    for NIFTI_file in NIFTI_files:
        basename = os.path.basename(NIFTI_file)
        shutil.move(NIFTI_file, os.path.join(outdir, basename))

    for folder in glob(os.path.join(outdir, "*")):
        if os.path.isdir(folder):
            shutil.rmtree(folder)

    return True


def download_project(
    xnat_url,
    datafolder,
    nsubjects="all",
    verbose=False,
    dataset="all",
):
    """Download data of full XNAT project."""
    # Connect to XNAT and retreive project
    labels_df_path = os.path.join(datafolder, "labels.csv")
    project_name = "worc"
    with xnat.connect(xnat_url) as session:
        project = session.projects[project_name]

        os.makedirs(datafolder, exist_ok=True)

        subjects_len = len(project.subjects)
        if nsubjects != "all":
            nsubjects = min(nsubjects, subjects_len)

        subjects_counter = 1
        downloaded_subjects_counter = 0
        labels = {}
        subjects = (
            name
            for name in project.subjects
            if project.subjects[name].fields["dataset"] == dataset
        )
        for subject_name in subjects:
            s = project.subjects[subject_name]
            if dataset != "all":
                # Check if patient belongs to required dataset
                subject_dataset = s.fields["dataset"]
                if subject_dataset != dataset:
                    continue

            # print(f"Processing on subject {subjects_counter}/{subjects_len}")
            subjects_counter += 1

            subject_diagnosis = int(s.fields["diagnosis_binary"])
            labels[s.label] = subject_diagnosis

            success = download_subject(
                project_name, s, datafolder, session, verbose
            )
            if success:
                downloaded_subjects_counter += 1

            # Stop downloading if we have reached the required number of subjects
            if downloaded_subjects_counter == nsubjects:
                break
        # Disconnect the session
        session.disconnect()
        if nsubjects != "all":
            if downloaded_subjects_counter < nsubjects:
                raise ValueError(
                    f"Number of subjects downloaded {downloaded_subjects_counter} is smaller than the number required {nsubjects}."
                )

        print("Done downloading!")

        label_df = pd.DataFrame(
            {
                "patient_ID": list(labels.keys()),
                "diagnosis": list(labels.values()),
            }
        )
        label_df.to_csv(labels_df_path, index=False)


def download_WORCDatabase(
    dataset=None,
    data_folder=None,
    n_subjects="all",
):
    """Download a dataset from the WORC Database.
    Download all Nifti images and segmentations from a dataset from the WORC
    database from https://xnat.bmia.nl/data/projects/worc
    dataset: string, default None
        If None, download the full XNAT project. If string, download one
        of the six datasets. Valid values: Lipo, Desmoid, GIST, Liver, CRLM,
        Melanoma
    """
    # Check if dataset is valid
    if dataset not in valid_datasets:
        raise KeyError(
            f"{dataset} is not a valid dataset, should be one of {valid_datasets}."
        )

    if data_folder is None:
        # Download data to path in which this script is located + Data
        cwd = os.getcwd()
        datafolder = os.path.join(cwd, "Data")
        if not os.path.exists(datafolder):
            os.makedirs(datafolder)

    xnat_url = "https://xnat.bmia.nl"
    download_project(
        xnat_url,
        data_folder,
        nsubjects=n_subjects,
        verbose=False,
        dataset=dataset,
    )



import datetime
import logging
import os
import time
from typing import Sequence

log = logging.getLogger(__name__)


def time_it(func):
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        log.info(func.__name__ + " took " + str(end - start) + "sec")
        return result

    return wrapper


def calculate_age(dob):
    """
    Calculate the age of a person from his date of birth.
    """
    today = datetime.datetime.now()
    return (
        today.year
        - dob.year
        - ((today.month, today.day) < (dob.month, dob.day))
    )


def calculate_age_at(date, dob):
    """
    Calculate the age of a person from his date of birth.

    """
    return (
        date.year - dob.year - ((date.month, date.day) < (dob.month, dob.day))
    )


def calculate_time_between(date1, date2):
    """
    Calculate the time between two dates.
    """
    return (date2 - date1).days


def filter_pyradiomics_names(names: Sequence[str]):
    """
    Filter features used in pyradiomics.
    """

    return [
        col
        for col in names
        if col.startswith(("original", "wavelet", "log-sigma"))
    ]


def set_n_jobs(n_jobs):
    """
    Set the number of parallel processes used by pyradiomics.
    """
    if n_jobs == -1:
        return os.cpu_count()
    else:
        return n_jobs



import logging
from pathlib import Path

import pandas as pd
from joblib import Parallel, delayed
from radiomics import featureextractor
from tqdm import tqdm

from autorad.config import config
from autorad.config.type_definitions import PathLike
from autorad.data.dataset import ImageDataset
from autorad.utils.utils import set_n_jobs, time_it

log = logging.getLogger(__name__)
# Silence the pyRadiomics logger
logging.getLogger("radiomics").setLevel(logging.WARNING)


class FeatureExtractor:
    def __init__(
        self,
        dataset: ImageDataset,
        feature_set: str = "pyradiomics",
        extraction_params: PathLike = "Baessler_CT.yaml",
        n_jobs: int | None = None,
    ):
        """
        Args:
            dataset: ImageDataset containing image paths, mask paths, and IDs
            feature_set: library to use features from (for now only pyradiomics)
            extraction_params: path to the JSON file containing the extraction
                parameters, or a string containing the name of the file in the
                default extraction parameter directory
                (autorad.config.pyradiomics_params)
            n_jobs: number of parallel jobs to run
        Returns:
            None
        """
        self.dataset = dataset
        self.feature_set = feature_set
        self.extraction_params = self._get_extraction_param_path(
            extraction_params
        )
        log.info(f"Using extraction params from {self.extraction_params}")
        self.n_jobs = set_n_jobs(n_jobs)
        self._initialize_extractor()

    def _get_extraction_param_path(self, extraction_params: PathLike) -> Path:
        default_extraction_param_dir = Path(config.PARAM_DIR)
        if Path(extraction_params).is_file():
            result = Path(extraction_params)
        elif (default_extraction_param_dir / str(extraction_params)).is_file():
            result = default_extraction_param_dir / extraction_params
        else:
            raise ValueError(
                f"Extraction parameter file {extraction_params} not found."
            )
        return result

    def run(self, keep_metadata=True) -> pd.DataFrame:
        """
        Run feature extraction.
        Args:
            keep_metadata: merge extracted features with data from the
            ImageDataset.df.
        """
        log.info("Extracting features")
        if self.n_jobs is None or self.n_jobs == 1:
            feature_df = self.get_features()
        else:
            feature_df = self.get_features_parallel()

        ID_colname = self.dataset.ID_colname
        feature_df = feature_df.astype(float)
        feature_df.insert(0, ID_colname, self.dataset.ids)

        if keep_metadata:
            # Add all columns from ImageDataset.df
            try:
                feature_df = self.dataset.df.merge(
                    feature_df,
                    left_on=ID_colname,
                    right_on=ID_colname,
                )
            except ValueError:
                raise ValueError("Error concatenating features and metadata.")
        return feature_df

    def save_config(self):
        """
        Save the extraction parameters to a JSON file. Should I use MLFlow here?
        """
        pass

    def _initialize_extractor(self):
        if self.feature_set == "pyradiomics":
            self.extractor = PyRadiomicsExtractorWrapper(
                str(self.extraction_params)
            )
        else:
            raise ValueError("Feature set not supported")
        log.info(f"Initialized extractor {self.feature_set}")
        return self

    def get_features_for_single_case(
        self, image_path: PathLike, mask_path: PathLike
    ) -> dict | None:
        """
        Returns:
            feature_series: dict with extracted features
        """
        if not Path(image_path).is_file():
            log.warning(
                f"Image not found. Skipping case... (path={image_path}"
            )
            return None
        if not Path(mask_path).is_file():
            log.warning(f"Mask not found. Skipping case... (path={mask_path}")
            return None
        try:
            feature_dict = self.extractor.execute(
                str(image_path),
                str(mask_path),
            )
        except ValueError:
            error_msg = f"Error extracting features for image, \
                mask pair {image_path}, {mask_path}"
            log.error(error_msg)
            return None

        return feature_dict

    @time_it
    def get_features(self) -> pd.DataFrame:
        """
        Get features for all cases.
        """
        lst_of_feature_dicts = [
            self.get_features_for_single_case(image_path, mask_path)
            for image_path, mask_path in tqdm(
                zip(self.dataset.image_paths, self.dataset.mask_paths)
            )
        ]
        feature_df = pd.DataFrame(lst_of_feature_dicts)
        return feature_df

    @time_it
    def get_features_parallel(self) -> pd.DataFrame:
        with Parallel(n_jobs=self.n_jobs) as parallel:
            lst_of_feature_dicts = parallel(
                delayed(self.get_features_for_single_case)(
                    image_path, mask_path
                )
                for image_path, mask_path in zip(
                    self.dataset.image_paths, self.dataset.mask_paths
                )
            )
        feature_df = pd.DataFrame(lst_of_feature_dicts)
        return feature_df

    def get_pyradiomics_feature_names(self) -> list[str]:
        class_obj = featureextractor.getFeatureClasses()
        feature_classes = list(class_obj.keys())
        feature_names = [
            f"{klass}_{name}"
            for klass in feature_classes
            for name in class_obj[klass].getFeatureNames().keys()
        ]
        return feature_names


class PyRadiomicsExtractorWrapper(featureextractor.RadiomicsFeatureExtractor):
    """Wrapper that filters out extracted metadata"""

    def __init__(self, extraction_params: PathLike, *args, **kwargs):
        super().__init__(str(extraction_params), *args, **kwargs)

    def execute(self, image_path: PathLike, mask_path: PathLike) -> dict:
        feature_dict = dict(super().execute(str(image_path), str(mask_path)))
        feature_dict_without_metadata = {
            feature_name: feature_dict[feature_name]
            for feature_name in feature_dict.keys()
            if "diagnostic" not in feature_name
        }
        return feature_dict_without_metadata



import itertools
import logging
from pathlib import Path
from typing import Sequence

import nibabel as nib
import numpy as np
import SimpleITK as sitk
from monai.transforms import (
    Compose,
    EnsureChannelFirstd,
    LoadImaged,
    ResampleToMatchd,
)

from autorad.config.type_definitions import PathLike

log = logging.getLogger(__name__)


def get_border_outside_mask_mm(
    mask_path: PathLike,
    margin: float | Sequence[float],
    output_path: PathLike,
):
    """Wrapper that takes in paths
    instead of sitk.Image.
    """
    mask = sitk.ReadImage(str(mask_path))
    border_mask = get_border_outside_mask_mm_sitk(mask, margin=margin)
    sitk.WriteImage(border_mask, str(output_path))


def get_border_outside_mask_mm_sitk(mask, margin: float | Sequence[float]):
    dilated_mask = dilate_mask_mm_sitk(mask, margin=margin)
    # subtract mask from dilated mask to get border
    border = dilated_mask - mask
    return border


def dilate_mask_mm(
    mask_path: PathLike,
    margin: float | Sequence[float],
    output_path: PathLike,
):
    """Wrapper that takes in paths
    instead of sitk.Image.
    """
    mask = sitk.ReadImage(str(mask_path))
    dilated_mask = dilate_mask_mm_sitk(mask, margin=margin)
    sitk.WriteImage(dilated_mask, str(output_path))


def dilate_mask_mm_sitk(mask, margin: float | Sequence[float]):
    """Dilate a mask in 3D by a margin given in mm.
    Accepts margin as a single float number or sequence of 3 floats."""
    if isinstance(margin, int):
        margins = np.array([margin, margin, margin])
    else:
        margins = np.array(margin)
    if margins.size != 3:
        raise ValueError("margin must be a float or a 3-tuple")
    spacing = np.array(mask.GetSpacing())
    if any((m > 0) and (m < s) for m, s in zip(margins, spacing)):
        raise ValueError(
            f"Margin = {margin} mm cannot be non-zero and smaller than the spacing = {spacing}"
        )
    dilation_in_voxels = margins / spacing
    dilation_in_voxels = tuple(
        int(np.round(voxel)) for voxel in dilation_in_voxels
    )
    real_dilation_in_mm = dilation_in_voxels * np.array(spacing)
    log.info(f"Dilating by a margin in mm: {real_dilation_in_mm}")
    dilated_mask = sitk.BinaryDilate(
        mask,
        dilation_in_voxels,
    )
    return dilated_mask


def center_of_mass(array: np.ndarray) -> list[float]:
    total = array.sum()
    ndim = len(array.shape)
    result = []
    for dim in range(ndim):
        other_dims = tuple(i for i in range(ndim) if i != dim)
        coord = (
            array.sum(axis=(other_dims)) @ range(array.shape[dim])
        ) / total
        result.append(coord)
    return result


def generate_bbox_around_mask_center(mask, bbox_size):
    """Mask a bounding box with fixed side length
    around the center of the input amask.
    """
    mask_center = center_of_mass(mask)
    mask_center = np.array(mask_center).round().astype(int)
    result = np.zeros_like(mask)
    margin_left = bbox_size // 2
    margin_right = bbox_size - margin_left
    low = mask_center - margin_left
    high = mask_center + margin_right
    result[
        max(0, low[0]) : min(mask.shape[0], high[0]),
        max(0, low[1]) : min(mask.shape[1], high[1]),
        max(0, low[2]) : min(mask.shape[2], high[2]),
    ] = 1
    return result


# taken from
# https://vincentblog.xyz/posts/medical-images-in-python-computed-tomography
def get_window(
    image: np.ndarray,
    window_center: float,
    window_width: float,
):
    img_min = window_center - window_width / 2
    img_max = window_center + window_width / 2
    window_image = image.copy()
    window_image[window_image < img_min] = img_min
    window_image[window_image > img_max] = img_max
    window_image_scaled = (window_image - img_min) / (img_max - img_min) * 255

    return window_image_scaled.astype(int)


# values taken from https://radiopaedia.org/articles/windowing-ct
def window_with_preset(image, window):
    if window == "soft tissues":
        return get_window(image, window_center=50, window_width=400)
    elif window == "bone":
        return get_window(image, window_center=400, window_width=1800)
    elif window == "lung":
        return get_window(image, window_center=-600, window_width=1500)
    elif window == "brain":
        return get_window(image, window_center=40, window_width=80)
    elif window == "liver":
        return get_window(image, window_center=30, window_width=150)
    else:
        raise ValueError(f"Unknown window setting: {window}")


def crop_volume_from_coords(coords_start, coords_end, vol):
    return vol[
        coords_start[0] : coords_end[0],
        coords_start[1] : coords_end[1],
        coords_start[2] : coords_end[2],
    ]


def get_largest_cross_section(mask, axis=2):
    if mask is None:
        raise ValueError("No mask loaded")
    ndims = len(mask.shape)
    other_axes = tuple(i for i in range(ndims) if i != axis)
    mask_sums = np.sum(mask, axis=other_axes)
    max_slicenum = np.argmax(mask_sums)
    return max_slicenum


def get_sitk_interpolator(interpolation):
    interpolation_mapper = {
        "nearest": sitk.sitkNearestNeighbor,
        "linear": sitk.sitkLinear,
        "bspline": sitk.sitkBSpline,
        "gaussian": sitk.sitkGaussian,
    }
    try:
        sitk_interpolator = interpolation_mapper[interpolation]
    except ValueError:
        raise ValueError(f"Interpolation {interpolation} not supported.")
    return sitk_interpolator


def resample_to_isotropic_sitk(
    image,
    interpolation="nearest",
    spacing=None,
    default_value=0,
    standardize_axes=False,
):
    """Implementation adapted from:
    https://github.com/InsightSoftwareConsortium/SimpleITK-Notebooks/
    blob/master/Python/05_Results_Visualization.ipynb
    """
    original_spacing = image.GetSpacing()
    # Image is already isotropic, just return a copy.
    if all(spc == original_spacing[0] for spc in original_spacing):
        return sitk.Image(image)
    # Make image isotropic via resampling.
    original_size = image.GetSize()
    if spacing is None:
        spacing = min(original_spacing)
    new_spacing = [spacing] * image.GetDimension()
    new_size = [
        int(round(osz * ospc / spacing))
        for osz, ospc in zip(original_size, original_spacing)
    ]
    interpolator = get_sitk_interpolator(interpolation)
    new_direction = image.GetDirection()
    new_origin = image.GetOrigin()
    # Only need to standardize axes if user requested and the original
    # axes were not standard.
    if standardize_axes and not np.array_equal(
        np.array(new_direction), np.identity(image.GetDimension()).ravel()
    ):
        new_direction = np.identity(image.GetDimension()).ravel()
        # Compute bounding box for the original, non standard axes image.
        boundary_points = []
        for boundary_index in list(
            itertools.product(*zip([0, 0, 0], image.GetSize()))
        ):
            boundary_points.append(
                image.TransformIndexToPhysicalPoint(boundary_index)
            )
        max_coords = np.max(boundary_points, axis=0)
        min_coords = np.min(boundary_points, axis=0)
        new_origin = min_coords
        new_size = (
            ((max_coords - min_coords) / spacing).round().astype(int)
        ).tolist()
    isotropic_img = sitk.Resample(
        image,
        new_size,
        sitk.Transform(),
        interpolator,
        new_origin,
        new_spacing,
        new_direction,
        default_value,
        image.GetPixelID(),
    )
    return isotropic_img


def resample_to_isotropic(img_path, output_path, interpolation="nearest"):
    """Wrapper for resample_to_isotropic_sitk that takes in paths
    instead of sitk.Image.
    """
    if output_path is None:
        output_path = img_path
    img = sitk.ReadImage(str(img_path))
    isotropic_img = resample_to_isotropic_sitk(
        image=img,
        interpolation=interpolation,
    )
    sitk.WriteImage(isotropic_img, str(output_path))


def load_and_resample_to_match(
    to_resample, reference, interpolation="nearest"
):
    """
    Args:
        to_resample: Path to the image to resample.
        reference: Path to the reference image.
    """
    data_dict = {"img": to_resample, "ref": reference}
    transforms = Compose(
        [
            LoadImaged(("img", "ref")),
            EnsureChannelFirstd(("img", "ref")),
            ResampleToMatchd("img", "ref_meta_dict", mode=interpolation),
        ]
    )
    result = transforms(data_dict)

    return result["img"][0], result["ref"][0]


def resample_to_img_sitk(img, target_img, interpolation="nearest"):
    """Resample image to target image.
    Both images should be sitk.Image instances.
    """
    interpolator = get_sitk_interpolator(interpolation)
    resampled_img = sitk.Resample(
        img,
        target_img,
        sitk.Transform(),
        interpolator,
        0,
        img.GetPixelID(),
    )
    return resampled_img


def resample_to_img(
    img_path, ref_path, output_path=None, interpolation="nearest"
):
    """
    Wrapper for resample_to_img_sitk that takes in paths instead of
    sitk.Image.
    """
    if output_path is None:
        output_path = img_path
    nifti = sitk.ReadImage(str(img_path))
    ref_nifti = sitk.ReadImage(str(ref_path))
    nifti_resampled = resample_to_img_sitk(
        img=nifti, target_img=ref_nifti, interpolation=interpolation
    )
    sitk.WriteImage(nifti_resampled, output_path)


def combine_nifti_masks(mask1_path, mask2_path, output_path):
    """
    Args:
        mask1_path: abs path to the first nifti mask
        mask2_path: abs path to the second nifti mask
        output_path: abs path to saved concatenated mask
    """
    if not Path(mask1_path).exists():
        raise FileNotFoundError(f"Mask {mask1_path} not found.")
    if not Path(mask2_path).exists():
        raise FileNotFoundError(f"Mask {mask2_path} not found.")

    mask1 = nib.load(mask1_path)
    mask2 = nib.load(mask2_path)

    matrix1 = mask1.get_fdata()
    matrix2 = mask2.get_fdata()
    assert matrix1.shape == matrix2.shape

    new_matrix = np.zeros(matrix1.shape)
    new_matrix[matrix1 == 1] = 1
    new_matrix[matrix2 == 1] = 2
    new_matrix = new_matrix.astype(int)

    new_mask = nib.Nifti1Image(
        new_matrix, affine=mask1.affine, header=mask1.header
    )
    nib.save(new_mask, output_path)


def relabel_mask(mask_path: str, label_map: dict[int, int], save_path):
    """
    Relabel mask with a new label map.
    E.g. for for a prostate mask with two labels:
    1 for peripheral zone and 2 for transition zone,
    relabel_mask(mask_path, {1: 1, 2: 1}) would merge both zones
    into label 1.
    """
    if not Path(mask_path).exists():
        raise FileNotFoundError(f"Mask {mask_path} not found.")
    mask = nib.load(mask_path)
    matrix = mask.get_fdata()
    n_found_labels = len(np.unique(matrix))
    if n_found_labels != len(label_map) + 1:
        raise ValueError(
            f"Number of unique labels in the mask is {n_found_labels}\
              and label map has {len(label_map)} items."
        )
    new_matrix = np.zeros(matrix.shape)
    for old_label, new_label in label_map.items():
        new_matrix[matrix == old_label] = new_label
    new_mask = nib.Nifti1Image(
        new_matrix, affine=mask.affine, header=mask.header
    )
    nib.save(new_mask, save_path)


def separate_nifti_masks(
    combined_mask_path, output_dir, label_dict=None, overwrite=False
):
    """
    Split multilabel nifti mask into separate binary nifti files.
    Args:
        combined_mask_path: abs path to the combined nifti mask
        output_dir: abs path to the directory to save separated masks
        label_dict: (optional) dictionary with names for each label
    """
    if not Path(combined_mask_path).exists():
        raise FileNotFoundError(f"Mask {combined_mask_path} not found.")

    mask = nib.load(combined_mask_path)
    matrix = mask.get_fdata()

    labels = np.unique(matrix).astype(int)
    labels = labels[labels != 0]

    assert sorted(labels), sorted(list(label_dict.keys()))

    for label in labels:
        new_matrix = np.zeros(matrix.shape)
        new_matrix[matrix == label] = 1
        new_matrix = new_matrix.astype(int)

        new_mask = nib.Nifti1Image(
            new_matrix, affine=mask.affine, header=mask.header
        )

        if label_dict is not None:
            label_name = label_dict[label]
        else:
            label_name = label

        output_path = Path(output_dir) / f"seg_{label_name}.nii.gz"
        if (not output_path.exists()) or (overwrite is True):
            nib.save(new_mask, output_path)



import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

import matplotlib.pyplot as plt
import pandas as pd
from sklearn.model_selection import train_test_split

from autorad.config import config
from autorad.config.type_definitions import PathLike
from autorad.utils import io, splitting, utils
from autorad.visualization import matplotlib_utils, plot_volumes

log = logging.getLogger(__name__)


@dataclass
class TrainingInput:
    train: pd.DataFrame
    test: pd.DataFrame
    val: Optional[pd.DataFrame] = None
    train_folds: Optional[list[pd.DataFrame]] = None
    val_folds: Optional[list[pd.DataFrame]] = None


@dataclass
class TrainingLabels:
    train: pd.Series
    test: pd.Series
    val: Optional[pd.DataFrame] = None
    train_folds: Optional[list[pd.DataFrame]] = None
    val_folds: Optional[list[pd.DataFrame]] = None


@dataclass
class TrainingMeta:
    train: pd.DataFrame
    test: pd.DataFrame
    val: Optional[pd.DataFrame] = None
    train_folds: Optional[list[pd.DataFrame]] = None
    val_folds: Optional[list[pd.DataFrame]] = None


@dataclass
class TrainingData:
    X: TrainingInput
    y: TrainingLabels
    meta: TrainingMeta

    _X_preprocessed: Optional[TrainingInput] = None
    _y_preprocessed: Optional[TrainingLabels] = None

    def iter_training(self):
        X, y = self.X_preprocessed, self.y_preprocessed
        if X is None:
            raise ValueError("No preprocessing done!")
        if X.val is not None:
            yield X.train, y.train, self.meta.train, X.val, y.val, self.meta.val
        elif X.train_folds is not None:
            yield from zip(
                X.train_folds,
                y.train_folds,
                self.meta.train_folds,
                X.val_folds,
                y.val_folds,
                self.meta.val_folds,
            )

    def __repr__(self):
        return f"TrainingData with {len(self.y.train)} training observations,\
        {len(self.y.test)} test observations, {self.X.train.shape[1]} features \
        and {self.meta.train.shape[1]} meta columns."

    @property
    def X_preprocessed(self):
        if self._X_preprocessed is None:
            raise ValueError("Preprocessing not performed!")
        return self._X_preprocessed

    @property
    def y_preprocessed(self):
        if self._y_preprocessed is None:
            raise ValueError("Preprocessing not performed!")
        return self._y_preprocessed

    @property
    def selected_features(self):
        if self._X_preprocessed is None:
            raise ValueError("Feature selection not performed!")
        return list(self._X_preprocessed.train.columns)


class FeatureDataset:
    """
    Store the extracted features and labels, split into training/test sets.
    """

    def __init__(
        self,
        dataframe: pd.DataFrame,
        target: str,
        ID_colname: str,
        features: Optional[list[str]] = None,
        meta_columns: list[str] = [],
        random_state: int = config.SEED,
    ):
        """
        Args:
            dataframe: table with extracted features
            target: name of the label column
            ID_colname: name of column with unique IDs for each case
            features: feature names
            meta_columns: columns to keep that are not features
            random_state: seed, used for splitting
        """
        self.df = dataframe
        self.target = target
        self.ID_colname = ID_colname
        self.random_state = random_state
        self.features = self._init_features(features)
        self.X: pd.DataFrame = self.df[self.features]
        self.y: pd.Series = self.df[self.target]
        self.meta_df = self.df[meta_columns + [ID_colname]]
        self._data: Optional[TrainingData] = None
        self.cv_splits: Optional[list[tuple[Any, Any]]] = None

    def _init_features(
        self, features: Optional[list[str]] = None
    ) -> list[str]:
        if features is None:
            all_cols = self.df.columns.tolist()
            features = utils.filter_pyradiomics_names(all_cols)
        return features

    @property
    def data(self):
        if self._data is None:
            raise AttributeError(
                "Data is not split into training/validation/test. \
                 Split the data or load splits from JSON."
            )
        else:
            return self._data

    def load_splits_from_json(self, json_path: PathLike, split_on=None):
        splits = io.load_json(json_path)
        return self.load_splits(splits, split_on)

    def load_splits(self, splits: dict, split_on=None):
        """
        `splits` dictionary should contain the following keys:
            - 'test': list of test IDs
            - 'train': dict with n keys (default n = 5)):
                - 'fold_{0..n-1}': list of training and
                                   list of validation IDs
        It can be created using `full_split()` defined below.
        """
        if split_on is None:
            split_on = self.ID_colname
        test_ids = splits["test"]
        test_rows = self.df[split_on].isin(test_ids)

        # Split dataframe rows
        X, y, meta = {}, {}, {}

        X["test"] = self.X.loc[test_rows]
        y["test"] = self.y.loc[test_rows]
        meta["test"] = self.meta_df.loc[test_rows]

        train_rows = ~self.df[split_on].isin(test_ids)

        if "val" in splits:
            val_ids = splits["val"]
            val_rows = self.df[split_on].isin(val_ids)
            train_rows = train_rows & ~val_rows

            X["val"] = self.X.loc[val_rows]
            y["val"] = self.y.loc[val_rows]
            meta["val"] = self.meta_df.loc[val_rows]

        X["train"] = self.X.loc[train_rows]
        y["train"] = self.y.loc[train_rows]
        meta["train"] = self.meta_df.loc[train_rows]

        if isinstance(splits["train"], dict):
            train_splits = splits["train"]
            n_splits = len(train_splits)
            self.cv_splits = [
                train_splits[f"fold_{i}"] for i in range(n_splits)
            ]
            X["train_folds"], X["val_folds"] = [], []
            y["train_folds"], y["val_folds"] = [], []
            meta["train_folds"], meta["val_folds"] = [], []
            for train_fold_ids, val_fold_ids in self.cv_splits:

                train_fold_rows = self.df[split_on].isin(train_fold_ids)
                val_fold_rows = self.df[split_on].isin(val_fold_ids)

                X["train_folds"].append(self.X[train_fold_rows])
                X["val_folds"].append(self.X[val_fold_rows])
                y["train_folds"].append(self.y[train_fold_rows])
                y["val_folds"].append(self.y[val_fold_rows])
                meta["train_folds"].append(self.meta_df[train_fold_rows])
                meta["val_folds"].append(self.meta_df[val_fold_rows])
        self._data = TrainingData(
            TrainingInput(**X), TrainingLabels(**y), TrainingMeta(**meta)
        )
        return self

    def split(
        self,
        save_path: Optional[PathLike] = None,
        method="train_with_cross_validation_test",
        split_on: Optional[str] = None,
        test_size: float = 0.2,
        *args,
        **kwargs,
    ):
        if split_on is None:
            split_on = self.ID_colname
        if method == "train_with_cross_validation_test":
            splits = self.full_split(split_on, test_size, *args, **kwargs)
        elif method == "train_val_test":
            splits = self.split_train_val_test(
                split_on, test_size, *args, **kwargs
            )
        else:
            raise ValueError(f"Method {method} is not supported.")

        if save_path is not None:
            Path(save_path).parent.mkdir(parents=True, exist_ok=True)
            io.save_json(splits, save_path)
            self.load_splits_from_json(save_path)
        return splits

    def full_split(
        self,
        split_on: str,
        test_size: float = 0.2,
        n_splits: int = 5,
    ):
        """
        Split into test and training, split training into k folds.
        """
        patient_df = self.df[[split_on, self.target]].drop_duplicates()
        if not patient_df[split_on].is_unique:
            raise ValueError(
                f"Selected column {split_on} has varying labels for the same ID!"
            )
        ids = patient_df[split_on].tolist()
        labels = patient_df[self.target].tolist()
        splits = splitting.split_full_dataset(
            ids=ids,
            labels=labels,
            test_size=test_size,
            n_splits=n_splits,
        )
        return splits

    def split_train_val_test(
        self,
        split_on: str,
        test_size: float = 0.2,
        val_size: float = 0.2,
    ):
        ids = self.df[split_on].tolist()
        labels = self.df[self.target].tolist()
        splits = splitting.split_train_val_test(
            ids=ids,
            labels=labels,
            test_size=test_size,
            val_size=val_size,
        )
        return splits

    def get_train_test_split_from_column(
        self, column_name: str, test_value: str
    ):
        """
        Use if the splits are already in the dataframe.
        """
        X_train = self.X[self.X[column_name] != test_value]
        y_train = self.y[self.y[column_name] != test_value]
        X_test = self.X[self.X[column_name] == test_value]
        y_test = self.y[self.y[column_name] == test_value]

        return X_train, y_train, X_test, y_test

    def split_train_val_test_from_column(
        self, column_name: str, test_value: str, val_size: float = 0.2
    ):
        data = {}
        (
            X_train_and_val,
            y_train_and_val,
            data["X_test"],
            data["y_test"],
        ) = self.get_train_test_split_from_column(column_name, test_value)
        (
            data["X_train"],
            data["X_val"],
            data["y_train"],
            data["y_val"],
        ) = train_test_split(
            X_train_and_val,
            y_train_and_val,
            test_size=val_size,
            random_state=self.random_state,
        )
        self._data = TrainingData(**data)

    def full_split_with_test_from_column(
        self,
        column_name: str,
        test_value: str,
        save_path: PathLike,
        split_on: Optional[str] = None,
        split_label: Optional[str] = None,
        n_splits: int = 5,
    ):
        """
        Splits into train and test according to `column_name`,
        then performs stratified k-fold cross validation split
        on the training set.
        """
        if split_on is None:
            split_on = self.ID_colname
        if split_label is None:
            split_label = self.target
        df_to_split = self.df[
            [split_on, split_label, column_name]
        ].drop_duplicates()
        if not df_to_split[split_on].is_unique:
            raise ValueError(
                f"Selected column {split_on} has varying labels for the same ID!"
            )
        train_to_split = df_to_split[df_to_split[column_name] != test_value]
        ids_train = train_to_split[split_on].tolist()
        y_train = train_to_split[split_label].tolist()
        ids_test = df_to_split.loc[
            df_to_split[column_name] == test_value, split_on
        ].tolist()

        ids_train_cv = splitting.split_cross_validation(
            ids_train, y_train, n_splits, random_state=self.random_state
        )

        ids_split = {
            "split_type": f"predefined test as {column_name} = {test_value}"
            " and stratified cross validation on training",
            "test": ids_test,
            "train": ids_train_cv,
        }
        io.save_json(ids_split, save_path)
        self.load_splits_from_json(save_path)

        return self


class ImageDataset:
    """
    Stores paths to the images, segmentations and labels
    """

    def __init__(
        self,
        df: pd.DataFrame,
        image_colname: str = "image_path",
        mask_colname: str = "segmentation_path",
        ID_colname: Optional[str] = None,
        root_dir: Optional[PathLike] = None,
    ):
        """
        Args:
            df: dataframe with image and mask paths
            image_colname: name of the image column in df
            mask_colname: name of the mask column in df
            ID_colname: name of the ID column in df. If None,
                IDs are assigned sequentially
            root_dir: root directory of the dataset, if needed
                to resolve paths
        """
        self._df = df
        self.image_colname = self._check_if_in_df(image_colname)
        self.mask_colname = self._check_if_in_df(mask_colname)
        self._set_ID_col(ID_colname)
        self.root_dir = Path(root_dir)

    def _check_if_in_df(self, colname: str):
        if colname not in self._df.columns:
            raise ValueError(
                f"{colname} not found in columns of the dataframe."
            )
        if self._df[colname].isnull().any():
            raise ValueError(f"{colname} contains null values")
        return colname

    def _set_new_IDs(self):
        log.info("ID not set. Assigning sequential IDs.")
        if "ID" not in self._df.columns:
            self.ID_colname = "ID"
        else:
            self.ID_colname = "ID_autogenerated"
        self._df[self.ID_colname] = range(len(self._df))

    def _set_ID_col_from_given(self, id_colname: str):
        if id_colname not in self._df.columns:
            raise ValueError(f"{id_colname} not in columns of dataframe.")
        ids = self._df[id_colname]
        # assert IDs are unique
        if len(ids.unique()) != len(ids):
            raise ValueError("IDs are not unique!")
        self.ID_colname = id_colname

    def _set_ID_col(self, id_colname: Optional[str] = None):
        if id_colname is None:
            self._set_new_IDs()
        else:
            self._set_ID_col_from_given(id_colname)

    @property
    def df(self) -> pd.DataFrame:
        """If root_dir is set, returns the dataframe with paths resolved"""
        if self.root_dir is None:
            return self._df
        result = self._df.copy()
        return result.assign(
            **{
                self.image_colname: self._df[self.image_colname].apply(
                    lambda x: os.path.join(self.root_dir, x)
                )
            }
        ).assign(
            **{
                self.mask_colname: self._df[self.mask_colname].apply(
                    lambda x: os.path.join(self.root_dir, x)
                )
            }
        )

    @property
    def image_paths(self) -> list[str]:
        return self.df[self.image_colname].to_list()

    @property
    def mask_paths(self) -> list[str]:
        return self.df[self.mask_colname].to_list()

    @property
    def ids(self) -> list[str]:
        return self.df[self.ID_colname].to_list()

    def plot_examples(self, n: int = 1, window="soft tissues"):
        if n > len(self.image_paths):
            n = len(self.image_paths)
            log.info(
                f"Not enough cases. Plotting all the cases instead (n={n})"
            )
        df_to_plot = self.df.sample(n)
        nrows, ncols, figsize = matplotlib_utils.get_subplots_dimensions(n)
        fig, axs = plt.subplots(nrows, ncols, figsize=figsize)
        for i in range(len(df_to_plot)):
            case = df_to_plot.iloc[i]
            ax = axs.flat[i]
            vols = plot_volumes.BaseVolumes.from_nifti(
                case[self.image_colname],
                case[self.mask_colname],
                window=window,
            )
            image_2D, mask_2D = vols.get_slices()
            single_plot = plot_volumes.overlay_mask_contour(image_2D, mask_2D)
            ax.imshow(single_plot)
            ax.set_title(f"{case[self.ID_colname]}")
            ax.axis("off")
        # return fig
